/*
 * Copyright (C) 2009-2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE
#define _REENTRANT

#include <pthread.h>

#include <sys/epoll.h>
#include <sys/time.h>
#include <errno.h>
#include <error.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define __LAIO_LIB__
#include "laio.h"

#define Abort(e, eno) error_at_line(e, eno, __FILE__, __LINE__, "%d %d", e, eno)

#define LaNthreadLimDef	20
#define LaNioLimDef	64

struct la_th {
	int stop_me;
	pthread_t th;
};

static struct {
	aio_context_t ctx;
	struct la_th *getevents;
	struct la_counter nthread, nio;

	struct {
		pthread_mutex_t mtx;
		pthread_cond_t cond;
	} suspend;
} laio = {
	.getevents	= NULL,

	/* values borrowed from glibc/sysdeps/pthread/aio_misc.c */
	.nthread	= {
		.mtx	= PTHREAD_MUTEX_INITIALIZER,
		.lim	= LaNthreadLimDef
	},

	.nio		= {
		.mtx	= PTHREAD_MUTEX_INITIALIZER,
		.lim	= LaNioLimDef
	},

	.suspend = {
		.mtx	= PTHREAD_MUTEX_INITIALIZER,
		.cond	= PTHREAD_COND_INITIALIZER
	}
};

#ifndef offsetof
#define offsetof(TYPE, MEMBER)  __builtin_offsetof (TYPE, MEMBER)
#endif

/* borrowed from linux/kernel.h */
/**
 * container_of - cast a member of a structure out to the containing structure
 * @ptr:	the pointer to the member.
 * @type:	the type of the container struct this is embedded in.
 * @member:	the name of the member within the struct.
 *
 */
#define container_of(ptr, type, member) ({			\
	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
	(type *)( (char *)__mptr - offsetof(type,member) );})

/* ---------------------------------------------------------------------- */

#define DPri(fmt, args...) \
	fprintf(stderr, "%s:%d:[%d]:" fmt, \
		__FILE__, __LINE__, getpid(), ##args)

/* ---------------------------------------------------------------------- */

static int la_cnt_read(struct la_counter *cnt)
{
	int v;

	pthread_mutex_lock(&cnt->mtx);
	v = cnt->cur;
	pthread_mutex_unlock(&cnt->mtx);

	return v;
}

static void la_cnt_add(struct la_counter *cnt, int n)
{
	pthread_mutex_lock(&cnt->mtx);
	cnt->cur += n;
	pthread_mutex_unlock(&cnt->mtx);
}

static int la_cnt_sub(struct la_counter *cnt, int n)
{
	int ret;

	pthread_mutex_lock(&cnt->mtx);
	cnt->cur -= n;
	ret = cnt->cur;
	pthread_mutex_unlock(&cnt->mtx);

	return ret;
}

static void la_cnt_inc(struct la_counter *cnt)
{
	pthread_mutex_lock(&cnt->mtx);
	cnt->cur++;
	pthread_mutex_unlock(&cnt->mtx);
}

static void la_cnt_dec(struct la_counter *cnt)
{
	pthread_mutex_lock(&cnt->mtx);
	cnt->cur--;
	pthread_mutex_unlock(&cnt->mtx);
}

/* ---------------------------------------------------------------------- */

static int la_create_thread(pthread_t *th, void *(*func)(void *),
			    pthread_attr_t *attr, void *arg)
{
	int err;
	pthread_attr_t _attr;

	err = 0;
	if (!attr) {
		attr = &_attr;
		err = pthread_attr_init(attr);
		if (!err)
			err = pthread_attr_setdetachstate
				(attr, PTHREAD_CREATE_DETACHED);
	}

	if (!err) {
		if (la_cnt_read(&laio.nthread) < laio.nthread.lim) {
			err = pthread_create(th, attr, func, arg);
			if (!err)
				la_cnt_inc(&laio.nthread);
		} else {
			err = -1;
			errno = EAGAIN;
		}
	}

	return err;
}

/* ---------------------------------------------------------------------- */

struct la_th_data {
	void (*func)(sigval_t v);
	sigval_t val;
};

static void *la_notify_thread(void *_arg)
{
	struct la_th_data arg, *p = _arg;

	arg = *p;
	free(p);
	arg.func(arg.val);
	la_cnt_dec(&laio.nthread);

	return NULL;
}

static int la_notify(pid_t pid, struct sigevent *se)
{
	int err;
	pthread_t th;
	struct la_th_data *arg;

	//DPri("se %p{v, no %d, notify %d}\n", se, se->sigev_signo, se->sigev_notify);
	err = 0;
	if (se->sigev_notify == SIGEV_SIGNAL
	    //|| se->sigev_notify == SIGEV_THREAD_ID
		)
		err = sigqueue(pid, se->sigev_signo, se->sigev_value);
	else if (se->sigev_notify == SIGEV_THREAD) {
		err = -1;
		arg = malloc(sizeof(*arg));
		if (!arg)
			goto out;

		arg->func = se->sigev_notify_function;
		arg->val = se->sigev_value;
		err = la_create_thread(&th, la_notify_thread,
				       se->sigev_notify_attributes, arg);
	}

 out:
	return err;
}

/*
 * set them only once since aio_return/error() can access them anytime
 */
static void la_fill(struct la_cb *lacb, int ret, int ec, int do_cond)
{
	/*
	 * set return value first,
	 * with hoping aio_error() will be called first
	 * and CPU executes so.
	 */
	if (do_cond) {
		if (ec == EINPROGRESS)
			Abort(-1, EINPROGRESS);
		pthread_mutex_lock(&laio.suspend.mtx);
		lacb->la_aiocb.__return_value = ret;
		lacb->la_aiocb.__error_code = ec;
		pthread_cond_broadcast(&laio.suspend.cond);
		pthread_mutex_unlock(&laio.suspend.mtx);
	} else {
		lacb->la_aiocb.__return_value = ret;
		lacb->la_aiocb.__error_code = ec;
	}
}

static int la_complete(struct la_cb *lacb, int res)
{
	int err;
	int (*sync)(int fd);
	unsigned long ul;

	if (res > 0)
		la_fill(lacb, res, 0, /*do_cond*/1);
	else if (res == 0) {
		/* dirtry trick, cf. laio_sync() */
		/* and dirty type cast */
		ul = (unsigned long)lacb->la_iocb.aio_data;
		sync = (void *)ul;
		if (!sync)
			la_fill(lacb, 0, 0, /*do_cond*/1);
		else {
			err = sync(lacb->la_aiocb.aio_fildes);
			if (!err)
				la_fill(lacb, 0, 0, /*do_cond*/1);
			else
				la_fill(lacb, err, errno, /*do_cond*/1);
		}
	} else if (res < 0)
		la_fill(lacb, -1, -res, /*do_cond*/1);
	la_cnt_dec(&laio.nio);

	if (lacb->la_lio && !la_cnt_sub(&lacb->la_lio->nio, 1)) {
		err = la_notify(lacb->la_pid, &lacb->la_lio->sev);
		//free(lacb->la_lio);
		//lacb->la_lio = NULL;
	} else
		err = la_notify(lacb->la_pid, &lacb->la_aiocb.aio_sigevent);
	return err;
}

static void *la_getevents(void *arg)
{
	int i, n, err;
	struct io_event ev[laio.nio.lim], *e;
	struct la_cb *lacb;
	struct iocb *iocb;
	struct la_th *lath = arg;

	while (!lath->stop_me) {
		//pthread_mutex_lock(&laio.suspend.mtx);
		n = io_getevents(laio.ctx, /*min_nr*/1, laio.nio.lim, ev,
				 /*timeout*/NULL);
		if (n < 0) {
			DPri("internal error, %d\n", errno);
			continue;
		}

		for (e = ev, i = 0; i < n; i++, e++) {
			//DPri("%llu, 0x%llx, %lld, %lld\n", e->data, e->obj, e->res, e->res2);
			lacb = container_of((void *)(unsigned long)e->obj,
					    struct la_cb, la_iocb);
			err = la_complete(lacb, e->res);
			if (err) {
				iocb = &lacb->la_iocb;
				DPri("AIO %d for fd %d completed "
				     "but notify failed, %d %d\n",
				     iocb->aio_lio_opcode, iocb->aio_fildes,
				     err, errno);
			}
		}
		//pthread_mutex_unlock(&laio.suspend.mtx);
	}
	free(lath);

	return NULL;
}

/* ---------------------------------------------------------------------- */

int laio_fin(void)
{
	int err;

	err = -1;
	errno = EINVAL;
	if (laio.ctx) {
		laio.getevents->stop_me = 1;
		pthread_cancel(laio.getevents->th);
		la_cnt_dec(&laio.nthread);
		laio.getevents = NULL;
		err = io_destroy(laio.ctx);
		laio.ctx = 0;
		laio.nthread.lim = LaNthreadLimDef;
		laio.nio.lim = LaNioLimDef;
	}

	return err;
}

static int la_init(int nthread, int nio)
{
	int err, e;
	static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;

	err = 0;
	if (laio.ctx)
		goto out;

	pthread_mutex_lock(&mtx);
	if (laio.ctx)
		goto out_unlock;

	err = -1;
	laio.nthread.lim = nthread;
	if (laio.nthread.lim < 0)
		laio.nthread.lim = 0;
	laio.nio.lim = nio;
	if (laio.nio.lim < 1)
		laio.nio.lim = 1;

	laio.getevents = calloc(1, sizeof(*laio.getevents));
	if (!laio.getevents)
		goto out_unlock;

	err = io_setup(laio.nio.lim, &laio.ctx);
	if (err)
		goto out_unlock;

	if (laio.nthread.lim)
		err = la_create_thread(&laio.getevents->th, la_getevents, NULL,
				       laio.getevents);

out_unlock:
	e = errno;
	pthread_mutex_unlock(&mtx);
	errno = e;
out:
	return err;
}

#ifdef __USE_GNU
void laio_init(const struct aioinit *init)
{
	int err;

	err = la_init(init->aio_threads, init->aio_num);
	if (err)
		Abort(err, errno); /* coz I am void */
}
#endif

/* ---------------------------------------------------------------------- */

#ifndef _ASM_GENERIC_SIGINFO_H
#define sigev_notify_thread_id	 _sigev_un._tid
#endif

static pid_t la_sev_pid(struct sigevent *sev)
{
	pid_t pid = 0;

	switch (sev->sigev_notify) {
	case SIGEV_SIGNAL:
		pid = getpid();
		break;
#if 0
	case SIGEV_THREAD_ID:
		pid = sev->sigev_notify_thread_id;
		break;
#endif
	}

	return pid;
}

static void la_iocb_init(struct la_cb *lacb, __u16 cmd, int do_set)
{
	struct iocb *iocb;
	unsigned long ul;

	iocb = &lacb->la_iocb;
	if (do_set) {
		ul = (unsigned long)lacb->la_aiocb.aio_buf;
		memset(iocb, 0, sizeof(*iocb));
		iocb->aio_buf = ul;
		iocb->aio_nbytes = lacb->la_aiocb.aio_nbytes;
		iocb->aio_offset = lacb->la_aiocb.aio_offset;
	}
	iocb->aio_lio_opcode = cmd;
	iocb->aio_fildes = lacb->la_aiocb.aio_fildes;
	la_fill(lacb, 0, EINPROGRESS, /*do_cond*/0);
	lacb->la_pid = la_sev_pid(&lacb->la_aiocb.aio_sigevent);
	lacb->la_lio = NULL;
}

static int la_rw(struct la_cb *lacb, __u16 cmd, int do_set)
{
	int err;
	struct iocb *iocb;

	err = la_init(laio.nthread.lim, laio.nio.lim);
	if (err)
		goto out;

	la_iocb_init(lacb, cmd, do_set);

	iocb = &lacb->la_iocb;
	err = io_submit(laio.ctx, /*nr*/1, &iocb);
	if (err == 1) {
		err = 0;
		la_cnt_inc(&laio.nio);
	} else
		la_fill(lacb, err, errno, /*do_cond*/0);

 out:
	return err;
}

int laio_read(struct la_cb *lacb)
{
	return la_rw(lacb, IOCB_CMD_PREAD, /*do_set*/1);
}

int laio_write(struct la_cb *lacb)
{
	return la_rw(lacb, IOCB_CMD_PWRITE, /*do_set*/1);
}

int laio_fsync(int op, struct la_cb *lacb)
{
	int err, flags;
	struct iocb *iocb;
	int (*sync)(int fd);

	err = -1;
	errno = EINVAL;
	if (!(op == O_SYNC || op == O_DSYNC))
		goto out;

	errno = EBADF;
	flags = fcntl(lacb->la_aiocb.aio_fildes, F_GETFL);
	if (flags == -1 || !(flags & (O_RDWR | O_WRONLY)))
		goto out;

	iocb = &lacb->la_iocb;
	if (flags & O_DIRECT) {
		/* nothing to do */
		err = la_init(laio.nthread.lim, laio.nio.lim);
		if (!err) {
			lacb->la_pid = la_sev_pid(&lacb->la_aiocb.aio_sigevent);
			lacb->la_lio = NULL;
			iocb->aio_data = 0;
			err = la_complete(lacb, 0);
		}
	} else {
		/*
		 * dirty trick to enque the aio sync request.
		 * Since there no FS which implements aio_sync,
		 * here we enque a write request with 0 bytes.
		 * the io_getevents thread will issue sync following ->aio_data
		 * after the write.
		 */
		memset(iocb, 0, sizeof(*iocb));
		sync = fsync;
		if (op == O_DSYNC)
			sync = fdatasync;
		iocb->aio_data = (unsigned long)sync;
		err = la_rw(lacb, IOCB_CMD_PWRITE, /*do_set*/0);
	}

out:
	return err;
}

ssize_t laio_return(struct la_cb *lacb)
{
	return lacb->la_aiocb.__return_value;
}

int laio_error(const struct la_cb *lacb)
{
	return lacb->la_aiocb.__error_code;
}

static int la_suspend_test(const struct la_cb * const lacb[], int n)
{
	int i;

	for (i = 0; i < n; i++)
		if (lacb[i]->la_aiocb.__error_code != EINPROGRESS)
			return 1;
	return 0;
}

int laio_suspend(const struct la_cb * const lacb[], int n,
		 const struct timespec *timeout)
{
	int err, e, real, i;
	struct timespec abs;
	const struct la_cb *placb[n];

	if (timeout) {
#if 0
		/* pthread_cond_timedwait() doesn't support MONOTONIC */
		err = clock_gettime(CLOCK_MONOTONIC, &abs);
		if (err)
			goto out;
		abs.tv_sec += timeout->tv_sec;
		abs.tv_nsec += timeout->tv_nsec;
#else
		/* convert the relative timeout into absolute */
		struct timeval now;

		gettimeofday(&now, NULL);
		abs.tv_nsec = timeout->tv_nsec + now.tv_usec * 1000;
		abs.tv_sec = timeout->tv_sec + now.tv_sec;
#endif
		if (abs.tv_nsec >= 1000 * 1000 * 1000) {
			abs.tv_nsec -= 1000 * 1000 * 1000;
			abs.tv_sec++;
		}
	}

	errno = EINVAL;
	err = -1;
	if (!laio.ctx || n > laio.nio.lim || n < 0)
		goto out;

	real = 0;
	for (i = 0; i < n; i++)
		if (lacb[i])
			placb[real++] = (struct la_cb *)lacb[i];

	err = 0;
	pthread_mutex_lock(&laio.suspend.mtx);
	while (!err && la_cnt_read(&laio.nio) > 0) {
		if (la_suspend_test(placb, real))
			break;

		if (!timeout)
			/* error never happen */
			pthread_cond_wait(&laio.suspend.cond,
					  &laio.suspend.mtx);
		else {
			err = pthread_cond_timedwait(&laio.suspend.cond,
						     &laio.suspend.mtx, &abs);
			if (err) {
				errno = err;
				err = -1;
			}
		}
	}
	e = errno;
	pthread_mutex_unlock(&laio.suspend.mtx);
	errno = e;

 out:
	if (errno == ETIMEDOUT)
		errno = EAGAIN;
	return err;
}

static int la_cancel_one(struct la_cb *lacb)
{
	int err, e;
	struct io_event ev;
	struct iocb *iocb;

	err = io_cancel(laio.ctx, &lacb->la_iocb, &ev);
	if (!err) {
		err = AIO_CANCELED;
		la_fill(lacb, -1, ECANCELED, /*do_cond*/1);
		la_cnt_dec(&laio.nio);
		e = la_notify(lacb->la_pid, &lacb->la_aiocb.aio_sigevent);
		if (e) {
			iocb = &lacb->la_iocb;
			DPri("AIO %d for fd %d successfully canceled "
			     "but notify failed, %d %d\n",
			     iocb->aio_lio_opcode, iocb->aio_fildes,
			     e, errno);
		}
	} else if (errno == EAGAIN)
		err = AIO_ALLDONE;
	else if (errno == EINVAL) /* non-Direct IO */
		err = AIO_NOTCANCELED;
	else
		Abort(err, errno);

	return err;
}

/*
 * Since io_cancel(2) doesn't distinguish the case of already completed,
 * this function doesn't return AIO_ALLDONE either.
 */
int laio_cancel(int fd, struct la_cb *lacb)
{
	int err /* , idx */;

	errno = EBADF;
	err = -1;
	if (fcntl(fd, F_GETFL) < 0)
		goto out;

	if (lacb) {
		errno = EINVAL;
		if (fd == lacb->la_aiocb.aio_fildes)
			err = la_cancel_one(lacb);
	} else {
		errno = ENOSYS;
#if 0
		/* unsupported */
		err = AIO_CANCELED;
		idx = 0;
		while (idx < laio.nio) {
			cb = la_cb_get_fd(fd, &idx);
			if (cb) {
				if (la_cancel_one(lacb))
					err = AIO_NOTCANCELED;
			} else
				break;
		}
#endif
	}

 out:
	return err;
}

static void la_wait(struct la_cb *_lacb)
{
	const struct la_cb * const lacb[] = {_lacb};

	if (la_suspend_test(lacb, 1))
		return;

	pthread_mutex_lock(&laio.suspend.mtx);
	while (1) {
		if (la_suspend_test(lacb, 1))
			break;
		pthread_cond_wait(&laio.suspend.cond, &laio.suspend.mtx);
	}
	pthread_mutex_unlock(&laio.suspend.mtx);
}

int llio_listio(int mode, struct la_cb * const lacb_arg[], int nent,
		struct sigevent *sig)
{
	int err, i, nio, cmd, e;
	struct la_cb *lacb, *placb[nent];
	pid_t pid;
	struct iocb *iocb[nent];
	struct la_lio *lio;

	err = -1;
	errno = EINVAL;
	if (mode == LIO_WAIT)
		sig = NULL;
	else if (mode != LIO_NOWAIT)
		goto out;

	pid = INT_MAX; /* hoping invalid value */
	lio = NULL;
	if (sig) {
		lio = malloc(sizeof(*lio));
		if (!lio)
			goto out;
		pthread_mutex_init(&lio->nio.mtx, NULL);
		lio->sev = *sig;
		pid = la_sev_pid(sig);
	}

	for (i = 0, nio = 0; i < nent; i++) {
		lacb = lacb_arg[i];
		if (!lacb)
			continue;

		switch (lacb->la_aiocb.aio_lio_opcode) {
		case LIO_READ:
			cmd = IOCB_CMD_PREAD;
			break;
		case LIO_WRITE:
			cmd = IOCB_CMD_PWRITE;
			break;
		case LIO_NOP:
			/* IOCB_CMD_NOOP is not implemented */
			continue; /* for loop */
		default:
			la_fill(lacb, -1, EINVAL, /*do_cond*/0);
			continue; /* for loop */
		}

		placb[nio] = lacb;
		iocb[nio] = &lacb->la_iocb;
		nio++;
		la_iocb_init(lacb, cmd, /*do_set*/1);
		lacb->la_pid = pid;
		lacb->la_lio = lio;
	}

	if (nio) {
		err = la_init(laio.nthread.lim, laio.nio.lim);
		if (err)
			goto out;

#if 1
		err = -1;
		errno = EINVAL;
		if (nio > (laio.nio.lim - la_cnt_read(&laio.nio)))
			goto out;
#endif

		if (lio)
			lio->nio.cur = nio;
		err = io_submit(laio.ctx, nio, iocb);
		if (err >= 0) {
			la_cnt_add(&laio.nio, err);
			if (err == nio)
				err = 0;
			else {
				/*
				 * io_submit() handles them sequentially
				 * and returns when the first error occurs.
				 */
				if (lio
				    && la_cnt_sub(&lio->nio, nio - err) <= 0) {
					e = la_notify(pid, sig);
					if (e)
						DPri("LIO completed "
						     "but notify failed, %d %d\n",
						     e, errno);
				}
				nio = err;
				err = -1;
				errno = EAGAIN;
				/* some are enqued */
			}
		} else
			goto out;

		if (mode == LIO_WAIT) {
			e = errno;
			for (i = 0; i < nio; i++)
				la_wait(placb[i]);
			errno = e;
		}
	} else if (sig) {
		e = la_notify(pid, sig);
		if (e)
			DPri("notify failed, %d %d\n", e, errno);
		err = 0;
	}

 out:
	return err;
}

/* ---------------------------------------------------------------------- */

#ifdef __LAIO_LIB64__
int laio_read64(struct la_cb64 *lacb)
{
	return laio_read(lacb);
}

int laio_write64(struct la_cb64 *lacb)
{
	return laio_write(lacb);
}

int laio_fsync64(int op, struct la_cb64 *lacb)
{
	return laio_fsync(op, lacb);
}

ssize_t laio_return64(struct la_cb64 *lacb)
{
	return laio_return(lacb);
}

int laio_error64(const struct la_cb64 *lacb)
{
	return laio_error(lacb);
}

int laio_suspend64(const struct la_cb64 * const lacb[], int n,
		   const struct timespec *timeout)
{
	return laio_suspend(lacb, n, timeout);
}

int laio_cancel64(int fd, struct la_cb64 *lacb)
{
	return laio_cancel(fd, lacb);
}

int llio_listio64(int mode, struct la_cb64 * const lacb[], int nent,
		  struct sigevent *sig)
{
	return llio_listio(mode, lacb, nent, sig);
}
#endif
