/*
 * Copyright (C) 2009-2010 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "sock.h"

/*
 * simple socket functions
 */

/* set or clear non-blocking mode */
int set_nonblock(int fd, int sw)
{
	int flags;

	flags = fcntl(fd, F_GETFL);
	if (sw)
		flags |= O_NONBLOCK;
	else
		flags &= ~O_NONBLOCK;
	return fcntl(fd, F_SETFL, flags);
}

/*
 * sockaddr as client
 * the result is stored in 'rsa'.
 * returns zero in success.
 */
static int simple_client_sa(struct sockaddr_in *rsa, char *host, int port)
{
	int err, herr;
	struct hostent he, *hp;
	char a[BUFSIZ];

	err = gethostbyname_r(host, &he, a, sizeof(a), &hp, &herr);
	if (err)
		goto out;

	memset(rsa, 0, sizeof(*rsa));
	memcpy(&rsa->sin_addr, hp->h_addr, hp->h_length);
	rsa->sin_family = hp->h_addrtype;
	rsa->sin_port = htons(port);

 out:
	return err;
}

/*
 * make a socket for client
 * returns the socket descriptor in success.
 */
static int simple_client_socket(char *host, int port, struct sockaddr_in *sa)
{
	int err, fd, v;

	err = simple_client_sa(sa, host, port);
	if (err)
		goto out;

	fd = socket(sa->sin_family, SOCK_STREAM, 0);
	if (fd < 0)
		goto out;

	v = 1;
	err = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &v, sizeof(v));
	if (!err)
		goto out; /* success */

	err = errno;
	close(fd);
	errno = err;
	fd = -1;
 out:
	return fd;
}

/*
 * make a socket for server
 * returns the socket descriptor in success.
 */
int simple_server_socket(int port, int backlog, int nonblock)
{
	int err, fd;
	struct sockaddr_in addr;

	fd = simple_client_socket("127.0.0.1", port, &addr);
	err = fd;
	if (fd < 0)
		goto out;

	if (nonblock) {
		err = set_nonblock(fd, nonblock);
		if (err)
			goto out_fd;
	}

	addr.sin_addr.s_addr = INADDR_ANY;
	err = bind(fd, (struct sockaddr *)&addr, sizeof(addr));
	if (err)
		goto out_fd;

	err = listen(fd, backlog);
	if (!err)
		goto out; /* success */

 out_fd:
	err = errno;
	close(fd);
	errno = err;
	fd = -1;
 out:
	return fd;
}

/* ---------------------------------------------------------------------- */

/* send/recv a socket descriptor between processes */

int send_sock(int pfd, int sock)
{
	int *p;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	char a[CMSG_SPACE(sizeof(int))];

	memset(&msg, 0, sizeof(msg));
	msg.msg_control = a;
	msg.msg_controllen = sizeof(a);
	cmsg = CMSG_FIRSTHDR(&msg);
	p = (void *)CMSG_DATA(cmsg);
	*p = sock;
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_RIGHTS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(int));
	msg.msg_controllen = cmsg->cmsg_len;

	return sendmsg(pfd, &msg, 0);
}

int recv_sock(int pfd)
{
	int err, *p;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	char a[CMSG_SPACE(sizeof(int))];

	memset(&msg, 0, sizeof(msg));
	msg.msg_control = a;
	msg.msg_controllen = sizeof(a);
	err = recvmsg(pfd, &msg, sizeof(msg));
	cmsg = CMSG_FIRSTHDR(&msg);
	if (!err
	    && cmsg->cmsg_level == SOL_SOCKET
	    && cmsg->cmsg_type == SCM_RIGHTS) {
		p = (void *)CMSG_DATA(cmsg);
		return *p;
	}
	return -1;
}
